/* Virtuino SE Json example:  Send - Receive text values
 * Supported boards: Arduino uno + Ethernet shield
 * Created by Ilias Lamprou
 * Updated Feb 27 2019
 * More examples: https://virtuino.com/index.php/virtuino-se/20-virtuino-se-code-examples
 */


#define outPin_1 6
#define inPin_1 7


#include <SPI.h>
#include <Ethernet.h>
//#include <Ethernet2.h>    // enable this line to support Ethernet 2
 
byte mac[] = {0xDE, 0xAD, 0xBE, 0xEF, 0xFE, 0xED};       // Set the ethernet shield mac address.
IPAddress ip(192, 168, 1, 150);                          // Set the ethernet shield ip address. Check your gateway ip address first
EthernetServer server(8000);                             // Server port =8000
    

//------------------- USER SETTINGS ---------------------------------
#define VIRTUINO_KEY  "1234"      // Virtuino KEY, only the requests that include this key will be acceptable. // disable the line if you don't want to use a key                          

#define V_memory_count 16         // the size of V memory. You can change it to a number <255)

String   textVariable1="";        // Add two String variables to store the incoming text from Virtuino SE
String   textVariable2="";
String   sensorText="";           // Add one String variable to store the text you want to send to Virtuino SE
int sensorValue=0;                // Add one int variable to store the value you want to send to Virtuino SE
int  reservedVariable1=0;       // reserverd variables
float  reservedVariable2=0;      

 
boolean debug = true;              // set this variable to false on the finale code to decrease the request time

//-------------------------------------------------------------------

                                  
unsigned long timeStrored=0;

//============================================================== setup
//==============================================================
void setup() {
  if (debug) {
    Serial.begin(9600);
    while (!Serial) continue;
  }
  Ethernet.begin(mac, ip);
  server.begin();
  if (debug) Serial.println(Ethernet.localIP());

  pinMode(outPin_1, OUTPUT);            
  pinMode(inPin_1, INPUT);
  
  }




//============================================================== loop
//==============================================================
void loop() {
  virtuinoRun();        //Necessary function to communicate with Virtuino. Client handler

  //--- How to send a text to Virtuino SE every time the state of a pin is changed
  if (digitalRead(inPin_1)==HIGH) textVariable2="The input pin is high";
  else textVariable2="The input pin is low";
  

  //--- How to read text variables from Virtuino SE.   It is better to move the next two lines to the function "onCommandReceived"
  if (textVariable1=="LED ON") digitalWrite(outPin_1,HIGH);             // Read the V2 memory.  On Virtuino panel add a regulator  to virtual memory V2  
     else if (textVariable1=="LED OFF")  digitalWrite(outPin_1,LOW);
  
  
  //--- How to send text and number values to Virtuino SE
   if (millis()-timeStrored>5000) {       // read sensor value every 5 seconds. Avoid to use delay on your code
      sensorValue = analogRead(A0);       // copy the A0 value to variable sensorValue
      sensorText = "The analog value="+String(sensorValue);
      timeStrored=millis();               // On Virtuino panel add TEXT Value displays virtual memories V2,V3
  }


}


//============================================================== onCommandReceived
//==============================================================
/* This function is called every time Virtuino SE app sends a request to server 
 * The 'variableIndex' is the V memory index of Virtuino app
 * The 'valueAsText' is the value that has sent from the app
 * If the 'valueAsText" is ? you have to return the value that you want to be displayed on Virtuino panel
 */
 String onCommandReceived(int variableIndex, String valueAsJson){     
   String valueAsText = getJsonValue(valueAsJson,"value");    // read the value from the received Json string 
   if (debug) Serial.println("New value received: V"+String(variableIndex)+"  value="+valueAsText);

    if (variableIndex==0){
       if (valueAsText=="?") return textVariable1;
       else textVariable1=valueAsText;
   }
   else if (variableIndex==1){
       if (valueAsText=="?") return textVariable2;
       else textVariable2=valueAsText;
   }
   else if (variableIndex==2){
       if (valueAsText=="?") return sensorText;
   }
  else if (variableIndex==3){
       if (valueAsText=="?") return String(sensorValue);
   }
   else if (variableIndex==4){
       if (valueAsText=="?") return String(reservedVariable1);
       else {
       int value = valueAsText.toInt();        // convert the value to int. The valueAsText have to be numerical
       reservedVariable1=value;
       }
   }

   else if (variableIndex==5){
       if (valueAsText=="?") return String(reservedVariable2);
       else {
        float value = valueAsText.toFloat();        // convert the value to float. The valueAsText have to be numerical
        reservedVariable2=value;
       }
   }
   
  return "";
}















//================== Virtuino Library (included in sketch) ===============
//==============================================================
//==============================================================
/* Only the next functions are nedded to communicate with Virtuino 
 * Avoid to make changes to the code below  
 */


#define WRONG_KEY_MESSAGE  "Wrong Key"  
#define WELLCOME_MESSAGE  "Hello Virtuino"  


//============================================================== virtuinoRun
//==============================================================
  void virtuinoRun(){
  EthernetClient client = server.available();
  if (client) {
    if (debug) Serial.println("Client connected");
    String data="";
    while (client.connected()) {
      while (client.available()>0) {
        char c = client.read();
        data+=c;
        if (debug) Serial.write(c);
      }
     if (debug) Serial.println("\nReceived data: "+data);
     String response= getResponseAsJson(data);
     if (debug) Serial.println("Response : "+response);
     client.print(response);
     client.stop();
    }
    if (debug) Serial.println("Client disconnected");
  }
 }



//============================================================== prepareResponse
//==============================================================
String getResponseAsJson(String data){
    //---check the KEY
   #ifdef VIRTUINO_KEY 
      String key = getJsonValue(data, "key");
      if (debug) Serial.println("Key="+key);
      if (!key.equals(VIRTUINO_KEY)) return "{\"status\":\"-1\",\"message\":\""+String(WRONG_KEY_MESSAGE)+"\"}";
   #endif

    //---check connection status
    String statusInfo = getJsonValue(data, "status");
    if (debug) Serial.println("statusInfo="+statusInfo);
    if (statusInfo.equals("0")) return "{\"status\":\"2\",\"message\":\""+String(WELLCOME_MESSAGE)+"\"}";
  
   String response="{\"status\":\"1\"";
   for (byte i=0;i<V_memory_count;i++){
      // easy way to read a Json object without installing extra json libraries
      String jsonTag="V"+String(i)+"_";
      int pos = data.indexOf(jsonTag);
      if (pos>0) {
        int startPos= data.indexOf("{",pos+1);
        if (startPos>pos) {
          int endPos= data.indexOf("}", startPos+2);
          if (endPos>startPos) {
            String variableAsJson = data.substring(startPos,endPos+1);
            String variableValue=onCommandReceived(i,variableAsJson);
           if (variableValue.length()>0) response+=",\""+jsonTag+"\":{\"value\":\""+urlencode(&variableValue)+"\"}";
          }
        }
      }
      
   } //for
   response+="}";
  return response;
}

//==============================================================
// easy way to read a Json text value without installing extra json libraries
  String getJsonValue(String jsonAsText, String jsonTag){
    int pos = jsonAsText.indexOf(jsonTag);
    if (pos>0) {
      int startPos=jsonAsText.indexOf(":\"",pos);
      int lastPos = jsonAsText.indexOf("\"",startPos+2);
      String value=jsonAsText.substring(startPos+2,lastPos);
      value=urldecode(&value);
      if (lastPos>startPos) return value;
    }
    return "";
 }


  //=================== ENCODE - DECODE Library (included in sketch) ==========
  //=================================================================
  
  //================================================================ urldecode
String urldecode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    for (int i =0; i < str->length(); i++){
        c=str->charAt(i);
      if (c == '+'){
        encodedString+=' ';  
      }else if (c == '%') {
        i++;
        code0=str->charAt(i);
        i++;
        code1=str->charAt(i);
        c = (h2int(code0) << 4) | h2int(code1);
        encodedString+=c;
      } else{
         encodedString+=c;  
      }
      yield();
    }
    
   return encodedString;
}
//================================================================ urlencode
String urlencode(String* str){
   String encodedString="";
    char c;
    char code0;
    char code1;
    char code2;
    for (int i =0; i < str->length(); i++){
      c=str->charAt(i);
      if (c == ' '){
        encodedString+= '+';
      } else if (isalnum(c)){
        encodedString+=c;
      } else{
        code1=(c & 0xf)+'0';
        if ((c & 0xf) >9){
            code1=(c & 0xf) - 10 + 'A';
        }
        c=(c>>4)&0xf;
        code0=c+'0';
        if (c > 9){
            code0=c - 10 + 'A';
        }
        code2='\0';
        encodedString+='%';
        encodedString+=code0;
        encodedString+=code1;
        //encodedString+=code2;
      }
      yield();
    }
    return encodedString;
}
//================================================================ h2int
unsigned char h2int(char c){
    if (c >= '0' && c <='9'){
        return((unsigned char)c - '0');
    }
    if (c >= 'a' && c <='f'){
        return((unsigned char)c - 'a' + 10);
    }
    if (c >= 'A' && c <='F'){
        return((unsigned char)c - 'A' + 10);
    }
    return(0);
}



